# Wrapper R function to call the calibart/calipred executables.


# pri.tree is the list created by maketreeprior()
# y is the simulator output
# yf is the field observation
# nu is for sigma and sigmaf prior
# sigq is quantile to use for sigma prior
# sigqf is quantile to use for sigmaf prior
# sigfhat is estimate of field observation measurement error
# kfac is used in variance of mean priors - max/min values of data are in +/1 kfac s.d.'s.
# rho is multiple used to shrink the variance of prior mean for the field data, idea
# is to force more of the signal to be taken up by better estimate of calibration parameter rather
# than the discrepancy.
makedatprior<-function(pri.tree,y,yf,nu,sigq,sigqf,sigfhat,kfac,rho,sighat=1e-5)
{
	m=pri.tree$m

	# sigma prior
	qchi = qchisq(1.0-sigq,nu)
	lambda = (sighat*sighat*qchi)/nu
	# sigmaf prior
	qchi = qchisq(1.0-sigqf,nu)
	lambdaf = (sigfhat*sigfhat*qchi)/nu

	# center the data
	my=mean(c(y,yf))
	y=y-my
	yf=yf-my

	maxy=max(y)
	miny=min(y)
	maxf=max(yf)
	minf=min(yf)
	maxyf=max(maxy,maxf)
	minyf=min(miny,minf)

	# prior sd of mean when tree ancestor has no Im split (ie mu0)
	tau0=(maxyf-minyf)/(2.0*kfac*sqrt(m))
	# prior mean of mean when tree ancestor has no Im spliit (ie mu0)
	mu0=mean(c(y,yf))/m
	# prior sd of mean when tree is left of Im split (ie muc == simulator data)
	tauc=(maxy-miny)/(2.0*kfac*sqrt(m))
	# prior mean of mean when tree is left of Im split (ie muc)
	muc=mean(y)/m
	# prior sd of mean when tree is right of Im split (ie mud == field data)
	taud=rho*(maxf-minf)/(2.0*kfac*sqrt(m))
	# prior mean of mean when tree is right of Im split (ie mud)
	mud=mean(yf)/m


	pri=list(nu=nu,nuf=nu,lambda=lambda,lambdaf=lambdaf,
			 tau0=tau0,mu0=mu0,tauc=tauc,muc=muc,taud=taud,mud=mud)
	return(pri)
}

# numcuts is the number of cutpoints for covariate and calibration parameter dimensions, typically 100-1000.
# alpha is for the depth prior which has the form alpha(1+depth)^(-beta).  Usually alpha=0.95.
# beta is for the depth prior, default is beta=2 which corresponds to tree depth around 2 or 3.  
#      Make beta smaller to make deeper trees more likely.
# m is the number of trees, typically 200 to 1000.
# pbirth is the probability of birth given birth/death.  Should be 0.5.
maketreeprior<-function(numcuts,alpha,beta,m=200,pbirth=0.5)
{
	pri=list(m=m,numcuts=numcuts,pbirth=pbirth,alpha=alpha,beta=beta)
	return(pri)	
}

# x is the design matrix of the simulator outputs
# caldims are the colums of x that are calibration parameters.
# nd is number of mcmc draws to keep/save as posterior draws.
# thin is to skip every "thin" MCMC draws after burn-in to limit storage requirements.
# burn is the number of initial burn-in draws, these are discarded.
# pbd is the probability of doing a birth/death proposal versus a rotation proposal.
# perturb.alpha is the scaling of the perturb proposal, typically something betwee 0.05 to 0.1.  This
# can be tuned to increase the tree acceptance rates if they are too low.
# ncores is number of cores to use with OpenMP.  Default is 4.
makemcmcinfo<-function(x,xf,y,yf,caldims,nd,thin,burn,pbd=0.8,perturb.alpha=0.05,ncores=4)
{
	x=as.matrix(cbind(x,0))
	imdim=ncol(x)
	xf=as.matrix(cbind(xf,1))

	# write out change-of-variable probability matrix
	ncaldims=length(caldims)
	cx=diag(imdim)
	cx[1:imdim,1:imdim]=1/imdim

	# center the data
	my=mean(c(y,yf))
	y=y-my
	yf=yf-my

	# rescale the x's
	r=matrix(0,nrow=imdim-1,ncol=2)
	for(i in 1:(imdim-1))
	{
		r[i,]=round(range(x[,i]))
		x[,i]=(x[,i]-r[i,1])/(r[i,2]-r[i,1])
		xf[,i]=(xf[,i]-r[i,1])/(r[i,2]-r[i,1])
	}
	for(i in 1:ncaldims) xf[,caldims[i]]=0.5

	mi=list(nd=nd,thin=thin,burn=burn,pbd=pbd,perturb.alpha=perturb.alpha,ncores=ncores,imdim=imdim,
		    caldims=caldims,ncaldims=ncaldims,cx=cx,x=x,xf=xf,y=y,yf=yf,my=my,r=r)
	return(mi)
}

# y is the output from the simulator
# yf is the field data
# x is the design matrix for simulator data, in the format [covariates|calibration params]
# xf is the design matrix for the observations, in the format [covariates]
# pri.dat is the prior information from makedatprior()
# pri.tree is the prior information from maketreeprior()
# mcmc.info is the mcmc information from makemcmcinfo()
calibrate<-function(pri.dat,pri.tree,mcmc.info)
{
	p=mcmc.info$imdim
	# write out transition proposal matrix
	write(t(mcmc.info$cx),ncol=p,file="chgvarmat.txt")
	# write out the simulator data
	write(mcmc.info$y,ncol=1,file=paste('y.txt',sep=""))
	write(t(mcmc.info$x),ncol=p,file=paste('x.txt',sep=""))
	# write out the field data
	write(mcmc.info$yf,ncol=1,file=paste('yf.txt',sep=""))
	write(t(mcmc.info$xf),ncol=p,file=paste('xf.txt',sep=""))
	# write out the calibration parameter dimensions
	write(mcmc.info$caldims,file="caldims.txt")
	# write out the model/field indicator variable Im
	write(mcmc.info$imdim,file="imdim.txt")

	# Now write out all the prior settings, etc, to settings file.
	parms=c("x","xf","y","yf",pri.dat$nu,pri.dat$nuf,pri.dat$lambda,pri.dat$lambdaf,
		    pri.dat$tau0,pri.dat$mu0,pri.dat$tauc,pri.dat$muc,pri.dat$taud,pri.dat$mud,
		    pri.tree$m,pri.tree$numcuts,pri.tree$pbirth,pri.tree$alpha,pri.tree$beta,
		    mcmc.info$nd,mcmc.info$thin,mcmc.info$burn,mcmc.info$pbd,mcmc.info$perturb.alpha)
	writeLines(parms,"settings.txt")

	cmd = paste('export OMP_NUM_THREADS=',mcmc.info$ncores,' ; ./calibart',sep='')
	time=system.time(system(cmd))
	cat("\n Runtime was ",time,"\n")

	# Unscale the results so we can return to user.
	# These are the in-sample posterior means for fitted field and simulator data.
	fit.yf=c(scan("bart-fit-field.txt"))
	fit.y=c(scan("bart-fit.txt"))
	fit.yf=fit.yf+mcmc.info$my
	fit.y=fit.y+mcmc.info$my

	# The draws of sigma, sigmaf
	fit.sd=c(scan("bart-sd.txt"))
	fit.sdf=c(scan("bart-sdf.txt"))

	# Calibration parameter estimates
	fit.thetas=as.matrix(read.table("bart-caldims.txt",header=F))
	# unscale them.
	for(i in 1:mcmc.info$ncaldims)
		fit.thetas[,i]=fit.thetas[,i]*(mcmc.info$r[caldims[i],2]-mcmc.info$r[caldims[i],1])+mcmc.info$r[caldims[i],1]

	return(list(yf=fit.yf,y=fit.y,sd=fit.sd,sdf=fit.sdf,thetas=fit.thetas,time=time))
}


# xp are the prediction points (covariates only)
# mcmc.info is from makemcmcinfo()
# requires calibrate() to have already successfully been called, and all the outputs of calibart
# are in the working directory.
# ql, qu are the lower and upper quantiles to report.  By default we also to the median.
predict<-function(xp,mcmc.info,ql,qu)
{
	# Make the columns for calibration params and Imdim
	xp=as.matrix(cbind(xp,0))
	np=nrow(xp)
	
	# Scale prediction points
	for(i in 1:(mcmc.info$imdim-1))
		if(!any(mcmc.info$caldims==i))
			xp[,i]=(xp[,i]-mcmc.info$r[i,1])/(mcmc.info$r[i,2]-mcmc.info$r[i,1])

	write(t(xp),ncol=ncol(xp),file=paste('xp.txt',sep=""))

	cmd = paste('export OMP_NUM_THREADS=',mcmc.info$ncores,' ;./calipred xp d',sep='')
	system(cmd)

	# Load the predictions
	bfitp=matrix(c(scan("yp.txt")),ncol=np,byrow=TRUE)  # model prediction
	bfitf=matrix(c(scan("yfp.txt")),ncol=np,byrow=TRUE) # field prediction
	bfitd=matrix(c(scan("yd.txt")),ncol=np,byrow=TRUE)  # discrepancy prediction
	# uncenter them
	bfitp=bfitp+mcmc.info$my
	bfitf=bfitf+mcmc.info$my
	bfitd=bfitd#+mcmc.info$my

	y.5=apply(bfitp,2,quantile,0.5)
	y.u=apply(bfitp,2,quantile,qu)
	y.l=apply(bfitp,2,quantile,ql)
	yf.5=apply(bfitf,2,quantile,0.5)
	yf.u=apply(bfitf,2,quantile,qu)
	yf.l=apply(bfitf,2,quantile,ql)
	d.5=apply(bfitd,2,quantile,0.5)
	d.u=apply(bfitd,2,quantile,qu)
	d.l=apply(bfitd,2,quantile,ql)

	return(list(y.5=y.5,y.u=y.u,y.l=y.l,yf.5=yf.5,yf.l=yf.l,yf.u=yf.u,d.5=d.5,d.u=d.u,d.l=d.l,
				posty=bfitp,postyf=bfitf,postd=bfitd))
}
