#include <iostream>

#include "info.h"
#include "tree.h"
#include "calibd.h"
#include "funs.h"
#include <algorithm>

using std::cout;
using std::endl;


bool calibd(tree& x, xinfo& xi, dinfo& di, pinfo& pi, RNG& gen)
{
   tree::npv goodbots;  //nodes we could birth at (split on)
   double PBx = getpb(x,xi,pi,goodbots); //prob of a birth at x
   unsigned int temp=0,temptsz=0;

	temptsz=goodbots[0]->depth();
	for(size_t i=1;i<goodbots.size();i++) {
		temp=goodbots[i]->depth();
		if(temp>temptsz)
			temptsz=temp;
	}
	pi.tproposal++;
	pi.tavgd+=temptsz;
	pi.tmaxd=std::max(pi.tmaxd,temptsz);

   if(gen.uniform() < PBx) { //do birth or death

      //--------------------------------------------------
      //draw proposal

      //draw bottom node, choose node index ni from list in goodbots 
      size_t ni = floor(gen.uniform()*goodbots.size()); 
      tree::tree_p nx = goodbots[ni]; //the bottom node we might birth at

      //draw v,  the variable
      std::vector<size_t> goodvars; //variables nx can split on
      getgoodvars(nx,xi,goodvars);
	  size_t v;
	  size_t vi = floor(gen.uniform()*goodvars.size()); //index of chosen split variable
      v = goodvars[vi];
      //draw c, the cutpoint
      int L,U;
      L=0; U = xi[v].size()-1;
      nx->rg(v,&L,&U);
      size_t c = L + floor(gen.uniform()*(U-L+1)); //U-L+1 is number of available split points

      //--------------------------------------------------
      //compute things needed for metropolis ratio

      double Pbotx = 1.0/goodbots.size(); //proposal prob of choosing nx
      size_t dnx = nx->depth();
      double PGnx = pi.alpha/pow(1.0 + dnx,pi.beta); //prior prob of growing at nx

      double PGly, PGry; //prior probs of growing at new children (l and r) of proposal
      if(goodvars.size()>1) { //know there are variables we could split l and r on
         PGly = pi.alpha/pow(1.0 + dnx+1.0,pi.beta); //depth of new nodes would be one more
         PGry = PGly;
      } else { //only had v to work with, if it is exhausted at either child need PG=0
         if((int)(c-1)<L) { //v exhausted in new left child l, new upper limit would be c-1
            PGly = 0.0;
         } else {
            PGly = pi.alpha/pow(1.0 + dnx+1.0,pi.beta);
         }
         if(U < (int)(c+1)) { //v exhausted in new right child r, new lower limit would be c+1
            PGry = 0.0;
         } else {
            PGry = pi.alpha/pow(1.0 + dnx+1.0,pi.beta);
         }
      }

      double PDy; //prob of proposing death at y
      if(goodbots.size()>1) { //can birth at y because splittable nodes left
         PDy = 1.0 - pi.pb;
      } else { //nx was the only node you could split on
         if((PGry==0) && (PGly==0)) { //cannot birth at y
            PDy=1.0;
         } else { //y can birth at either l or r
            PDy = 1.0 - pi.pb;
         }
      }

      double Pnogy; //death prob of choosing the nog node at y
      size_t nnogs = x.nnogs();
      tree::tree_cp nxp = nx->getp();
      if(nxp==0) { //no parent, nx is the top and only node
         Pnogy=1.0;
      } else {
         if(nxp->isnog()) { //if parent is a nog, number of nogs same at x and y
            Pnogy = 1.0/nnogs;
         } else { //if parent is not a nog, y has one more nog.
           Pnogy = 1.0/(nnogs+1.0);
         }
      }
  
      //--------------------------------------------------
      //compute sufficient statistics
      sinfo sl,sr; //sl for left from nx and sr for right from nx (using rule (v,c))
	  getCalisuff(x,nx,v,c,xi,di,sl,sr);

      //--------------------------------------------------
      //compute alpha

      double alpha=0.0,alpha1=0.0,alpha2=0.0;
      double lill=0.0,lilr=0.0,lilt=0.0;
      if( (sl.n+sl.nf)>=minperbot && (sr.n+sr.nf)>=minperbot ) {
      		calc_all_lil(nx, v, &lill, &lilr, &lilt, sl, sr, pi);

		alpha1 = (PGnx*(1.0-PGly)*(1.0-PGry)*PDy*Pnogy)/((1.0-PGnx)*PBx*Pbotx);
		alpha2 = alpha1*exp(lill+lilr-lilt);
		alpha = std::min(1.0,alpha2);
      }
	  else {
         alpha=0.0;
      }

      /*
      cout << "sigma, tau: " << pi.sigma << ", " << pi.tau << endl;
      cout << "birth prop: node, v, c: " << nx->nid() << ", " << v << ", " << c << "," << xi[v][c] << endl;
      cout << "L,U: " << L << "," << U << endl;
      cout << "PBx, PGnx, PGly, PGry, PDy, Pnogy,Pbotx:" <<
         PBx << "," << PGnx << "," << PGly << "," << PGry << "," << PDy <<
         ", " << Pnogy << "," << Pbotx << endl;
      cout << "left ss: " << sl.n << ", " << sl.sy << ", " << sl.sy2 << endl;
      cout << "right ss: " << sr.n << ", " << sr.sy << ", " << sr.sy2 << endl;
      cout << "lill, lilr, lilt: " << lill << ", " << lilr << ", " << lilt << endl;
      cout << "alphas: " << alpha1 << ", " << alpha2 << ", " << alpha << endl;
      */

      //--------------------------------------------------
      //finally ready to try metrop
      double a,ac,ad,b,bf,s2,sf2,yb,ybf;
      double mul,mur; //means for new bottom nodes, left and right
      if(gen.uniform() < alpha) {
	      pi.taccept++;

         //draw mul, mean for left node
         a= 1.0/(pi.tau0*pi.tau0);
         ac=1.0/(pi.tauc*pi.tauc);
         ad=1.0/(pi.taud*pi.taud);
         s2 = pi.sigma*pi.sigma;
	      sf2= pi.sigmaf*pi.sigmaf;

	 // draw mul for left node
	 if(nx->isrightchildofvsplit(pi.imdim)) {
		bf = sl.nf/sf2;
		ybf = sl.sf/sl.nf;
		mul = (bf*ybf+pi.mud*ad)/(ad+bf) + gen.normal()/sqrt(ad+bf);
	 }
    else if(nx->isleftchildofvsplit(pi.imdim) || v==pi.imdim)
    {
		b = sl.n/s2;
		yb = sl.sy/sl.n;
		mul = (b*yb + pi.muc*ac)/(ac+b) + gen.normal()/sqrt(ac+b);
	 }
	 else
    {
	 	b = sl.n/s2;
		yb = sl.sy/s2;
      bf = sl.nf/sf2;
      ybf = sl.sf/sf2;
		mul = (yb + ybf + pi.mu0*a)/(a+b+bf) + gen.normal()/sqrt(a+b+bf);
	 }

	 //draw mur for right node
	 if(nx->isrightchildofvsplit(pi.imdim) || v==pi.imdim) {
		bf = sr.nf/sf2;
		ybf = sr.sf/sr.nf;
		mur = (bf*ybf+pi.mud*ad)/(ad+bf) + gen.normal()/sqrt(ad+bf);
	 }
	 else if(nx->isleftchildofvsplit(pi.imdim)) {
		b = sr.n/s2;
		yb = sr.sy/sr.n;
      mur = (b*yb + pi.muc*ac)/(ac+b) + gen.normal()/sqrt(ac+b);
	 }
	 else {
		b = sr.n/s2;
		yb = sr.sy/s2;
      bf = sr.nf/sf2;
      ybf = sr.sf/sf2;
      mur = (yb + ybf + pi.mu0*a)/(a+b+bf) + gen.normal()/sqrt(a+b+bf);
	 }


         //do birth
	 //cout << "birth, mul=" << mul << " mur=" << mur << endl;
	 x.birth(nx->nid(),v,c,mul,mur);
         return true;
      } else {
         return false;
      }
   } else {
      //--------------------------------------------------
      //draw proposal

      //draw nog node, any nog node is a possibility
      tree::npv nognds; //nog nodes
      x.getnogs(nognds);
      size_t ni = floor(gen.uniform()*nognds.size()); 
      tree::tree_p nx = nognds[ni]; //the nog node we might kill children at

      //--------------------------------------------------
      //compute things needed for metropolis ratio

      double PGny; //prob the nog node grows
      size_t dny = nx->depth();
      PGny = pi.alpha/pow(1.0+dny,pi.beta);

      //better way to code these two?
      double PGlx = pgrow(nx->getl(),xi,pi);
      double PGrx = pgrow(nx->getr(),xi,pi);

      double PBy;  //prob of birth move at y
      //if(nx->ntype()=='t') { //is the nog node nx the top node
      if(!(nx->p)) { //is the nog node nx the top node
         PBy = 1.0;
      } else {
         PBy = pi.pb;
      }

      double Pboty;  //prob of choosing the nog as bot to split on when y
      int ngood = goodbots.size();
      if(cansplit(nx->getl(),xi)) --ngood; //if can split at left child, lose this one 
      if(cansplit(nx->getr(),xi)) --ngood; //if can split at right child, lose this one
      ++ngood;  //know you can split at nx
      Pboty=1.0/ngood;

      double PDx = 1.0-PBx; //prob of a death step at x
      double Pnogx = 1.0/nognds.size();

      //--------------------------------------------------
      //compute sufficient statistics
      sinfo sl,sr; //sl for left from nx and sr for right from nx (using rule (v,c))
	  getCalisuff(x,nx->getl(),nx->getr(),xi,di,sl,sr);

      //--------------------------------------------------
      //compute alpha
	  double lill=0.0, lilr=0.0, lilt=0.0;
      	calc_all_lil(nx, nx->getv(), &lill, &lilr, &lilt, sl, sr, pi);


		double alpha1 = ((1.0-PGny)*PBy*Pboty)/(PGny*(1.0-PGlx)*(1.0-PGrx)*PDx*Pnogx);
		double alpha2 = alpha1*exp(lilt - lill - lilr);
		double alpha = std::min(1.0,alpha2);

      /*
      cout << "death prop: " << nx->nid() << endl;
      cout << "nognds.size(), ni, nx: " << nognds.size() << ", " << ni << ", " << nx << endl;
      cout << "depth of nog node: " << dny << endl;
      cout << "PGny: " << PGny << endl;
      cout << "PGlx: " << PGlx << endl;
      cout << "PGrx: " << PGrx << endl;
      cout << "PBy: " << PBy << endl;
      cout << "Pboty: " << Pboty << endl;
      cout << "PDx: " << PDx << endl;
      cout << "Pnogx: " << Pnogx << endl;
      cout << "left ss: " << sl.n << ", " << sl.sy << ", " << sl.sy2 << endl;
      cout << "right ss: " << sr.n << ", " << sr.sy << ", " << sr.sy2 << endl;
      cout << "lill, lilr, lilt: " << lill << ", " << lilr << ", " << lilt << endl;
      cout << "sigma: " << pi.sigma << endl;
      cout << "tau: " << pi.tau << endl;
      cout << "alphas: " << alpha1 << ", " << alpha2 << ", " << alpha << endl;
      */

      //--------------------------------------------------
      //finally ready to try metrop
      double a,ac,ad,b,bf,s2,sf2,yb,ybf;
      double mu;
      size_t n,nf;
      if(gen.uniform()<alpha) {
         pi.taccept++;

         //draw mu for nog (which will be bot)
         n = sl.n + sr.n;
         nf= sl.nf + sr.nf;
         a= 1.0/(pi.tau0*pi.tau0);
         ac=1.0/(pi.tauc*pi.tauc); 
         ad=1.0/(pi.taud*pi.taud);
         s2 = pi.sigma*pi.sigma;
         sf2= pi.sigmaf*pi.sigmaf;

	 //draw mu for nog (which will be bot)
	 if(nx->isrightchildofvsplit(pi.imdim)) {
		bf = nf/sf2;
		ybf = (sl.sf+sr.sf)/nf;
		mu = (bf*ybf+pi.mud*ad)/(ad+bf) + gen.normal()/sqrt(ad+bf);
	 }
	 else if(nx->isleftchildofvsplit(pi.imdim)) {
		b = n/s2;
		yb = (sl.sy+sr.sy)/n;
		mu = (b*yb + pi.muc*ac)/(ac+b) + gen.normal()/sqrt(ac+b);
	 }
	 else {
		b = n/s2;
		yb = (sl.sy+sr.sy)/s2;
      bf = nf/sf2;
      ybf = (sl.sf+sr.sf)/sf2;
		mu = (yb + ybf + pi.mu0*a)/(a+b+bf) + gen.normal()/sqrt(a+b+bf);
	 }

         //do death
	 x.death(nx->nid(),mu);
         return true;
      } else {
         return false;
      }
   }
}


void calichgv(tree& x, xinfo& xi, dinfo& di, pinfo& pi, RNG& gen)
{
	tree::tree_p pertnode;
	unsigned int pertdx;
	std::vector<size_t> tryvars;
	size_t oldvar,newvar,oldcut,newcut;
	std::vector<std::vector<double> > cov;
	bool didswap=false;


	if(x.treesize()==1)
		return;

	// Get nog nodes and propose new split value
	tree::npv nognds;
	x.getnobots(nognds);  //This is all internal nodes of the tree.
	for(size_t k=0;k<nognds.size();k++) {
		if(gen.uniform()<0.5) //50% of the time we do change of variable, otherwise perturb.
		{
			pi.tproposal++;
			pertdx=k;
			pertnode = nognds[pertdx];

			int L,U;
			double oldcutprob=1.0, newcutprob=1.0;

			getLU(pertnode,xi,&L,&U);
			oldcutprob=((double)(U-L+1.0));

			cov=pi.cvpm; // a copy of the prior.  Start with the prior at every node of the tree...
			didswap=false;
			oldvar=pertnode->getv();
			oldcut=pertnode->getc();
			updatecormat(pertnode,xi,cov);
			normchgvrow(oldvar,cov);

			//choose new variable randomly
			newvar=getchgv(oldvar,cov,gen);
			pertnode->setv(newvar);
			if(cov[oldvar][newvar]<0.0) {
				pertnode->swaplr();
				didswap=true;
			}

			//get L,U for the new variable, save it and set newcut.
			int Ln,Un;
			size_t newcut;
			getLU(pertnode,xi,&Ln,&Un);
			newcutprob=((double)(Un-Ln+1.0));
			newcut=Ln+(size_t)(floor(gen.uniform()*(Un-Ln+1.0)));
			pertnode->setc(newcut);

	        //now we also need to update the row of chgv for newv->oldv to calc MH correctly
	        updatecormat(pertnode,xi,cov);
	        normchgvrow(newvar,cov);

	        //sanity check:
	        if(cov[newvar][oldvar]==0.0)
	           cout << "Proposal newv cannot return to oldv!  This is not possible!" << endl;
	        double alpha0=cov[newvar][oldvar]/cov[oldvar][newvar];  //proposal ratio for newvar->oldvar and oldvar->newvar


	        // get suff stats and do accept/reject
			std::vector<sinfo> svold;
			std::vector<sinfo> svnew;
			tree::npv bnv;
			getcalipertsuff(x,pertnode,bnv,oldcut,oldvar,didswap,xi,di,svold,svnew);

			typedef tree::npv::size_type bvsz;
			double lilold,lilnew;
			bool hardreject=false;
			lilold=0.0;
			for(bvsz j=0;j!=svold.size();j++) {

				if(bnv[j]->isrightchildofvsplit(pi.imdim))
					lilold += clil(0,svold[j].nf,0.0,svold[j].sf,pi.sigma,pi.sigmaf,pi.taud*pi.taud,pi.mud);
				else if(bnv[j]->isleftchildofvsplit(pi.imdim))
					lilold+= clil(svold[j].n,0,svold[j].sy,0.0,pi.sigma,pi.sigmaf,pi.tauc*pi.tauc,pi.muc);
				else
					lilold+= clil(svold[j].n,svold[j].nf,svold[j].sy,svold[j].sf,pi.sigma,pi.sigmaf,pi.tau0*pi.tau0,pi.mu0);
			}

			lilnew=0.0;
			for(bvsz j=0;j!=svnew.size();j++) {
				if( (svnew[j].n+svnew[j].nf) < minperbot)
					hardreject=true;

				if(bnv[j]->isrightchildofvsplit(pi.imdim))
					lilnew += clil(0,svnew[j].nf,0.0,svnew[j].sf,pi.sigma,pi.sigmaf,pi.taud*pi.taud,pi.mud);
				else if(bnv[j]->isleftchildofvsplit(pi.imdim))
					lilnew+= clil(svnew[j].n,0,svnew[j].sy,0.0,pi.sigma,pi.sigmaf,pi.tauc*pi.tauc,pi.muc);
				else
					lilnew+= clil(svnew[j].n,svnew[j].nf,svnew[j].sy,svnew[j].sf,pi.sigma,pi.sigmaf,pi.tau0*pi.tau0,pi.mu0);
			}

			double alpha1 = oldcutprob/newcutprob; //anything from the prior?
			double alpha2=alpha0*alpha1*exp(lilnew-lilold);
			double alpha = std::min(1.0,alpha2);

			if(hardreject) alpha=0.0;  //change of variable led to an bottom node with <5 observations in it, we reject this.

			if(gen.uniform()<alpha) {
				pi.taccept++;
				pertnode->setv(newvar); //because the call to getpertsuff changes it back to oldvar to calc old lil.
				pertnode->setc(newcut); //because the call to getpertsuff changes it back to oldcut to calc the old lil.
				if(didswap)
					pertnode->swaplr();
			}
			//else nothing, pertnode->c and pertnode->v is already reset to the old value and if a swap was done it is already unswapped.
		}
		else  //do a perturb proposal
		{
			didswap=false;
			pi.tproposal++;
			pertdx=k;
			pertnode = nognds[pertdx];
			int L,U;
			double oldcutprob=1.0, newcutprob=1.0;

			// Get allowable range for perturbing cv at pertnode
			getLU(pertnode,xi,&L,&U);
			newcutprob=1.0/(U-L+1.0);

			size_t oldc = pertnode->getc();
			size_t propc;
			int ai,bi,oldai,oldbi;
			ai=(int)(floor(oldc-pi.pertalpha*(U-L+1.0)/2.0));
			bi=(int)(floor(oldc+pi.pertalpha*(U-L+1.0)/2.0));
			ai=std::max(ai,L);
			bi=std::min(bi,U);
			propc = ai + (size_t)(floor(gen.uniform()*(bi-ai+1.0)));
			pertnode->setc(propc);
			oldai=(int)(floor(propc-pi.pertalpha*(U-L+1.0)/2.0));
			oldbi=(int)(floor(propc+pi.pertalpha*(U-L+1.0)/2.0));
			oldai=std::max(oldai,L);
			oldbi=std::min(oldbi,U);
			newcutprob=1.0/(bi-ai+1.0);
			oldcutprob=1.0/(oldbi-oldai+1.0);

			std::vector<sinfo> svold;
			std::vector<sinfo> svnew;
			tree::npv bnv;
			getcalipertsuff(x,pertnode,bnv,oldc,pertnode->getv(),didswap,xi,di,svold,svnew);

			typedef tree::npv::size_type bvsz;
			double lilold,lilnew;
			bool hardreject=false;
			lilold=0.0;
			for(bvsz j=0;j!=svold.size();j++) {

				if(bnv[j]->isrightchildofvsplit(pi.imdim))
					lilold += clil(0,svold[j].nf,0.0,svold[j].sf,pi.sigma,pi.sigmaf,pi.taud*pi.taud,pi.mud);
				else if(bnv[j]->isleftchildofvsplit(pi.imdim))
					lilold+= clil(svold[j].n,0,svold[j].sy,0.0,pi.sigma,pi.sigmaf,pi.tauc*pi.tauc,pi.muc);
				else
					lilold+= clil(svold[j].n,svold[j].nf,svold[j].sy,svold[j].sf,pi.sigma,pi.sigmaf,pi.tau0*pi.tau0,pi.mu0);
			}

			lilnew=0.0;
			for(bvsz j=0;j!=svnew.size();j++) {
				if( (svnew[j].n+svnew[j].nf) < minperbot)
					hardreject=true;

				if(bnv[j]->isrightchildofvsplit(pi.imdim))
					lilnew += clil(0,svnew[j].nf,0.0,svnew[j].sf,pi.sigma,pi.sigmaf,pi.taud*pi.taud,pi.mud);
				else if(bnv[j]->isleftchildofvsplit(pi.imdim))
					lilnew+= clil(svnew[j].n,0,svnew[j].sy,0.0,pi.sigma,pi.sigmaf,pi.tauc*pi.tauc,pi.muc);
				else
					lilnew+= clil(svnew[j].n,svnew[j].nf,svnew[j].sy,svnew[j].sf,pi.sigma,pi.sigmaf,pi.tau0*pi.tau0,pi.mu0);
			}

			double alpha1 = oldcutprob/newcutprob; //anything from the prior?
			double alpha2= alpha1*exp(lilnew-lilold);
			double alpha = std::min(1.0,alpha2);

			if(hardreject) alpha=0.0;  //perturb led to an bottom node with <5 observations in it, we reject this.

			if(gen.uniform()<alpha) {
				pi.taccept++;
				pertnode->setc(propc); //because the call to getpertsuff changes it back to oldc to calc the old lil.
			}
			//else nothing, pertnode->c and pertnode->v is already reset to the old value and if a swap was done it is already unswapped.	
		}
	}
}

//--------------------------------------------------
// Functions to support change-of-variable proposal
//--------------------------------------------------
// update the correlation matrix for chgv move taking into account that not all
// variables may be eligible at pertnode.
void updatecormat(tree::tree_p pertnode, xinfo& xi, std::vector<std::vector<double> >& chgv)
{
   int Ln,Un; //L,U for the ``new'' variable
   size_t oldv=pertnode->getv();
   size_t p=chgv.size();

   for(size_t i=0;i<p;i++) {
      if(i!=oldv && std::abs(chgv[oldv][i])>0.0) {
         if(chgv[oldv][i]<0.0)  //swap left,right branches
            pertnode->swaplr();
         getpertLU(pertnode,i,xi,&Ln,&Un);
         if(chgv[oldv][i]<0.0)  //undo the swap
            pertnode->swaplr();
         if(Un<Ln) //we can't transition to variable i here according to the tree structure
            chgv[oldv][i]=0.0;
      }
   }
}
//--------------------------------------------------
// renormalize the correlation matrix so that the probability of row sums to 1.
void normchgvrow(size_t row, std::vector<std::vector<double> >& chgv)
{
   double tmp=0.0;
   size_t p=chgv.size();

   for(size_t i=0;i<p;i++)
      tmp+=std::abs(chgv[row][i]);
   for(size_t i=0;i<p;i++)
      chgv[row][i]/=tmp;
}
//--------------------------------------------------
// randomly choose a new variable to transition to from oldv
size_t getchgv(size_t oldv, std::vector<std::vector<double> >& chgv, RNG& gen)
{
   double cp=gen.uniform();
   size_t p=chgv.size();
   size_t newv=oldv;
   std::vector<double> cumprob;

   cumprob=chgv[oldv];
   cumprob[1]=std::abs(cumprob[1]);
   for(size_t i=1;i<p;i++)
      cumprob[i]=std::abs(cumprob[i])+cumprob[i-1];

   for(size_t i=0;i<p;i++) {
      if(cumprob[i] >= cp) { //propose transition to this variable
         newv=i;
         i=p;  //break out of loop early
      }        
   }
   return newv;
}


