// funs.h: supporting functions for the MCMC
//         - suff stats for gibbs draws
//         - suff stats for MH steps
//         - serial and parallel (MPI) codepaths
//--------------------------------------------------
#include <cmath>
#include <iostream>
#include "tree.h"
#include "info.h"
#include "rng.h" 

using std::cout;
using std::endl;

//pi and log(2*pi)
#define PI 3.1415926535897931
#define LTPI 1.83787706640934536

//--------------------------------------------------
// Normal density
double pn(double x, double m, double v);

//--------------------------------------------------
// Draw from a discrete distribution
int rdisc(double *p, RNG& gen);

//--------------------------------------------------
// Evaluate tree tr on grid xi, write to os
void grm(tree& tr, xinfo& xi, std::ostream& os);

//--------------------------------------------------
// Does a (bottom) node have variables you can split on?
bool cansplit(tree::tree_p n, xinfo& xi);

//--------------------------------------------------
// Compute prob of a birth, goodbots will contain all the good bottom nodes
double getpb(tree& t, xinfo& xi, pinfo& pi, tree::npv& goodbots);

//--------------------------------------------------
// Find variables n can split on, put their indices in goodvars
void getgoodvars(tree::tree_p n, xinfo& xi, std::vector<size_t>& goodvars);
void getinternalvars(tree::tree_p n, xinfo& xi,  std::vector<size_t>& goodvars);
int getnumcuts(tree::tree_p n, xinfo& xi, size_t var);
bool isImgoodvar(tree::tree_p n, xinfo& xi);

//--------------------------------------------------
// Get prob a node grows, 0 if no good vars, else a/(1+d)^b
double pgrow(tree::tree_p n, xinfo& xi, pinfo& pi);


//--------------------------------------------------
// Get L,U for variable that pertnode splits on given the entire tree
void getLU(tree::tree_p pertnode, xinfo& xi, int* L, int* U);
void getpertLU(tree::tree_p pertnode, size_t pertvar, xinfo& xi, int* L, int* U);

//--------------------------------------------------
// Get sufficients stats for all bottom nodes
void allsuff(tree& x, xinfo& xi, dinfo& di, tree::npv& bnv, std::vector<sinfo>& sv);

//--------------------------------------------------
// Get sufficient stats for children (v,c) of node nx in tree x
void getsuff(tree& x, tree::tree_cp nx, size_t v, size_t c, xinfo& xi, dinfo& di, sinfo& sl, sinfo& sr);
void getpertsuff(tree& x, tree::tree_p pertnode, tree::npv& bnv, size_t oldc, xinfo& xi, dinfo& di, std::vector<sinfo>& svold, std::vector<sinfo>& svnew);
void getpertsuff(tree& x, tree::tree_p pertnode, tree::npv& bnv, size_t oldc, size_t oldvar, bool didswaplr, xinfo& xi, dinfo& di, std::vector<sinfo>& svold, std::vector<sinfo>& svnew);

//--------------------------------------------------
// Get sufficient stats for pair of bottom children nl(left) and nr(right) in tree x
void getsuff(tree& x, tree::tree_cp nl, tree::tree_cp nr, xinfo& xi, dinfo& di, sinfo& sl, sinfo& sr);

//--------------------------------------------------
#ifdef CALIBART
//get sufficient stats for children (v,c) of node nx in tree x for the CALIBART calibration model
void getCalisuff(tree& x, tree::tree_cp nx, size_t v, size_t c, xinfo& xi, dinfo& di, sinfo& sl, sinfo& sr);

//--------------------------------------------------
//get sufficient stats for pair of bottom children nl(left) and nr(right) in tree x for the CALIBART calibration model
void getCalisuff(tree& x, tree::tree_cp nl, tree::tree_cp nr, xinfo& xi, dinfo& di, sinfo& sl, sinfo& sr);


//--------------------------------------------------
// get all the suffs
void caliallsuff(tree& x, xinfo& xi, dinfo& di, tree::npv& bnv, std::vector<sinfo>& sv);

//--------------------------------------------------
// get the suffs for pert move
void getcalipertsuff(tree& x, tree::tree_p pertnode, tree::npv& bnv, size_t oldc, xinfo& xi, dinfo& di,
                     std::vector<sinfo>& svold, std::vector<sinfo>& svnew);
void getcalipertsuff(tree& x, tree::tree_p pertnode, tree::npv& bnv, size_t oldc, size_t oldv, bool didswaplr, xinfo& xi, dinfo& di,
                     std::vector<sinfo>& svold, std::vector<sinfo>& svnew);

//--------------------------------------------------
// draw the bottom node mu's
void calidrmu(tree& t, xinfo& xi, dinfo& di, pinfo& pi, RNG& gen);

//--------------------------------------------------
// draw the calibration parameters
void calidrcp(std::vector<tree>& t, xinfo& xi, dinfo& di, pinfo& pi,
              std::vector<double> &xf, std::vector<double> &xft, std::vector<double> &yf, double *allfitf, std::vector<double> &y, double *allfit, RNG& gen);
#endif

//--------------------------------------------------
// Log of the integreted likelihood
double lil(size_t n, double sy, double sy2, double sigma, double tau);

//--------------------------------------------------
#ifdef CALIBART
//log of the integreted likelihood for the calibart model
void calc_all_lil(tree::tree_p nx, size_t v, double* lill, double* lilr, double* lilt, sinfo& sl, sinfo& sr, pinfo& pi);
double clil(size_t n, size_t nf, double sy, double sf, double sigma, double sigmaf, double t2, double mu0);
#endif

//--------------------------------------------------
//fit
void fit(tree& t, xinfo& xi, dinfo& di, std::vector<double>& fv);
//--------------------------------------------------
//fit
void fit(tree& t, xinfo& xi, dinfo& di, double* fv);
//--------------------------------------------------
//fit
double fitone(tree& t,xinfo& xi, dinfo& di, size_t ix);
//--------------------------------------------------
//partition
void partition(tree& t, xinfo& xi, dinfo& di, std::vector<size_t>& pv);

//--------------------------------------------------
// Draw all the bottom node mu's
void drmu(tree& t, xinfo& xi, dinfo& di, pinfo& pi, RNG& gen);

//--------------------------------------------------
// Write cutpoint information to screen
void prxi(xinfo& xi);

//--------------------------------------------------
// Make xinfo = cutpoints
void makexinfo(size_t p, size_t n, double *x, xinfo& xi, size_t nc);

//--------------------------------------------------
// Get min/max for p predictors needed to make cutpoints.
void makeminmax(size_t p, size_t n, double *x, std::vector<double> &minx, std::vector<double> &maxx);

//--------------------------------------------------
void makexinfominmax(size_t p, xinfo& xi, size_t nc, std::vector<double> &minx, std::vector<double> &maxx);
// version with variable number of cutpoints per dimension:
void makexinfominmax(size_t p, xinfo& xi, std::vector<size_t> &nc, std::vector<double> &minx, std::vector<double> &maxx);

#ifdef CALIBART
//--------------------------------------------------
// make field and simulator predictions.
void makefieldpred(dinfo di, xinfo &xi, std::vector<tree> &t, double *fpred);
void makesimpred(dinfo di, xinfo& xi, std::vector<tree> &t, double *ypred);
void fitfield(tree& t, xinfo& xi, dinfo& di, double* fv);
#endif

