# Bayesian Regression Tree Calibration Model - Example.
# Matthew T. Pratola
# 2014.
#
# This is the ``small'' Nevada dataset example.
#
# R wrapper functions for calling the model, doing preditions, etc.
source("../calibart.r")



#################################################################################
# Setup the Nevada dataset example
#################################################################################
ncores=8      # number of parallel cpu cores to use with OpenMP
n=17600       # number of simulator outputs
nf=10         # number of field observations
np=100 		  # number of prediction points
caltrues=rep(.5,6)  # the "unknown" truth
caldims=3:8   # which columns of design matrix X are calibration parameters?
              # Note: currently we *require* the last columns of the matrix
              #       to be all calibration parameters.
p=244		  # total number of variables is covariates + calibration params + imdim = 244
# extract data:
datcmd='rm trees.txt; rm bart-*.txt; rm y.txt; rm yf.txt; rm x.txt; rm xf.txt; rm caldims.txt; rm imdim.txt; rm chgvarmat.txt; unzip ntsdata.zip'
system(datcmd)
X=matrix(c(scan("x.txt")),ncol=p,byrow=TRUE)
Xf=matrix(c(scan("xf.txt")),ncol=p,byrow=TRUE)
y=scan("y.txt")
yf=scan("yf.txt")
cvars=c(28,37,99,104,158,211)+2
X=X[,c(1,2,cvars)]
Xf=Xf[,c(1,2,cvars)]

#################################################################################
# Setup calibart model settings.
#################################################################################
numcuts=100   # number of cutpoints for (cts) x-dimensions and theta-dimensions
m=200         # number of trees in the sum-of-trees.
rho=0.01      # shrink the variance of mu|Im=1 by this factor from the "default"
sigfhat=0.05  # a priori guess that 10% of the field data is "noise"

# Setup MCMC settings
nd=5000      # number of draws to sample from the posterior after burn-in
thin=10      # reduce the number of draws we save to disk to nd/thin
burn="10000"     # number of initial draws discarded as burn-in, not enough but it will run in ~30mins.
pbd=0.7            # do birth/death proposal pbd% of the time, otherwise do rotate
perturb.alpha=0.02 # scaling parameter for perturb proposal, usually between
                   # 0.02-0.10, use this to tune acceptance rate of tree moves.
ncores=8      # number of parallel cpu cores when using OpenMP

# Other settings where defaults are usually fine.
alpha=0.95  # for the tree
beta=2.0    # depth prior.
nu=3        # \
sigq=0.99   #  for the
sigqf=0.90  #  variance priors.
kfac=2      # /

# Make the prior datastructures
tree.pri=maketreeprior(numcuts,alpha,beta,m)
dat.pri=makedatprior(tree.pri,y,yf,nu,sigq,sigqf,sigfhat,kfac,rho,sighat=0.07)
mcmc.info=makemcmcinfo(X,Xf,y,yf,caldims,nd,thin,burn,
	      pbd=pbd,perturb.alpha=perturb.alpha,ncores=ncores)


#################################################################################
# Run the model
#################################################################################
fit=calibrate(dat.pri,tree.pri,mcmc.info)

# Predict the out-of-sample field
Xff=matrix(0,nrow=100,ncol=8)
Xff[,8]=1
Xff[1:50,1]=0
Xff[51:100,1]=1
Xff[1:50,2]=seq(0,1,length=50)
Xff[51:100,2]=seq(0,1,length=50)
fitp=predict(Xff,mcmc.info,0.025,0.975)


#################################################################################
# Do some plots
#################################################################################
png("mcmc.png",width=600,height=600)
par(mfrow=c(2,2))
# In-sample fits
plot(yf,fit$yf,pch=20,cex=.5,xlab="Field Observations",ylab="Posterior Mean")
abline(0,1)
plot(y,fit$y,pch=20,cex=.5,xlab="Simulator Outputs", ylab="Posterior Mean")
abline(0,1)
# Trace plot of sigma,sigmaf
plot(fit$sd,type='l',xlab="Iteration",ylab=expression(sigma))
plot(fit$sdf,type='l',xlab="Iteration",ylab=expression(phi))
dev.off()

# Plot of calibration parameter estimates
png("mcmc-thetas.png",width=900,height=600)
par(mfrow=c(2,3))
plot(density(fit$thetas[,1]),xlim=c(0,1),xlab=expression(theta[28]),main="",lwd=2)
plot(density(fit$thetas[,2]),xlim=c(0,1),xlab=expression(theta[37]),main="",lwd=2)
plot(density(fit$thetas[,3]),xlim=c(0,1),xlab=expression(theta[99]),main="",lwd=2)
plot(density(fit$thetas[,4]),xlim=c(0,1),xlab=expression(theta[104]),main="",lwd=2)
plot(density(fit$thetas[,5]),xlim=c(0,1),xlab=expression(theta[158]),main="",lwd=2)
plot(density(fit$thetas[,6]),xlim=c(0,1),xlab=expression(theta[211]),main="",lwd=2)
dev.off()

# Plot of field predictions overlaid over the simulator outputs.  Red dots are the field observations.
ix0=which(X[,1]==0)
ix1=which(X[,1]==1)
if0=which(Xf[,1]==0)
if1=which(Xf[,1]==1)
fp0=1:50
fp1=51:100
pdf("fieldpred.pdf",width=12,height=6)
par(mfrow=c(1,2))
plot(X[ix0,2],y[ix0],pch=20,col="grey",xlab="time (normalized)",ylab="head (site 0)",ylim=c(-2,2))
lines(Xff[fp0,2],fitp$yf.5[fp0],col="green",lwd=2)
lines(Xff[fp0,2],fitp$yf.u[fp0],col="green",lty=2,lwd=2)
lines(Xff[fp0,2],fitp$yf.l[fp0],col="green",lty=2,lwd=2)
lines(Xff[fp0,2],fitp$d.5[fp0],col="orange",lwd=1)
lines(Xff[fp0,2],fitp$d.l[fp0],col="orange",lty=2,lwd=1)
lines(Xff[fp0,2],fitp$d.u[fp0],col="orange",lty=2,lwd=1)
points(Xf[if0,2],yf[if0],col="red",pch=20)
plot(X[ix1,2],y[ix1],pch=20,col="grey",xlab="time (normalized)",ylab="head (site 1)",ylim=c(-2,2))
lines(Xff[fp1,2],fitp$yf.5[fp1],col="green",lwd=2)
lines(Xff[fp1,2],fitp$yf.u[fp1],col="green",lty=2,lwd=2)
lines(Xff[fp1,2],fitp$yf.l[fp1],col="green",lty=2,lwd=2)
lines(Xff[fp1,2],fitp$d.5[fp1],col="orange",lwd=1)
lines(Xff[fp1,2],fitp$d.l[fp1],col="orange",lty=2,lwd=1)
lines(Xff[fp1,2],fitp$d.u[fp1],col="orange",lty=2,lwd=1)
points(Xf[if1,2],yf[if1],col="red",pch=20)
dev.off()


