
# Bayesian Regression Tree Calibration Model - Example.
# Matthew T. Pratola
# 2014.
#
# This is Tom's example.
# The simulation model is y(x,t1,t2)=t1*x + t2
# The field model is yf(x)=a*x^2 -x + 0.4 + epsilon 
#                 or yf(x)=a     -x + 0.4 + epsilon
# ie theta1=-1,theta2=0.4, bias=a*x^2 or a  (quadratic or constant)

# R wrapper functions for calling the model, doing preditions, etc.
source("../calibart.r")


set.seed(0)

#################################################################################
# Setup a Tom's example dataset.
#################################################################################
ncores=4      # number of parallel cpu cores to use with OpenMP
n=500         # number of simulator outputs
nf=25         # number of field observations
np=100 		  # number of prediction points
caltrues=c(-1,0.4)  # the "unknown" truth
sigmaf=0.1    # the true s.d. used to generate field data
caldims=2:3   # which columns of design matrix X are calibration parameters?
              # Note: currently we *require* the last columns of the matrix 
              #       to be all calibration parameters.

# Simulator data
X=matrix(runif(n*3,-5,5),ncol=3)  # design matrix for simulator data
y=X[,2]*X[,1]+X[,3]               # simulator outputs

# Field data
#biasalpha=2.0 # for constant discrepancy, 0=no bias
biasalpha=0.1 # for quadratic discrepancy, 0=no bias
# Could change the type of bias (discrepancy) here.
biasfn<-function(X,a)
{
#	return(a*rep(1,nrow(X)))	  #constant discrepancy
	return(a*X[,1]^2)                 #quadratic discrepancy
}
Xf=matrix(runif(nf*3,-5,5),ncol=3) # design matrix for field data
Xf[,caldims[1]]=caltrues[1]
Xf[,caldims[2]]=caltrues[2]
yf=biasfn(Xf,biasalpha)+caltrues[1]*Xf[,1]+caltrues[2]+rnorm(nf,sd=sigmaf)

# Prediction data
Xp=matrix(runif(3*np,-5,5),ncol=3) #prediction points for field
Xp[,caldims[1]]=caltrues[1]
Xp[,caldims[2]]=caltrues[2]
bias=biasfn(Xp,biasalpha) # This is the true bias at the prediction grid points
fp=bias+caltrues[1]*Xp[,1]+caltrues[2]  # This is the true field process at the 
                                        # prediction grid points



#################################################################################
# Setup calibart model settings.
#################################################################################
numcuts=1000  # number of cutpoints for (cts) x-dimensions and theta-dimensions
m=200         # number of trees in the sum-of-trees.
rho=0.05      # shrink the variance of mu|Im=1 by this factor from the "default"
sigfhat=sd(yf)*0.1  # a priori guess that 10% of the field data is "noise"

# Setup MCMC settings
nd=2000      # number of draws to sample from the posterior after burn-in
thin=10      # reduce the number of draws we save to disk to nd/thin
burn=4000    # number of initial draws discarded as burn-in
pbd=0.7      # do birth/death proposal pbd% of the time, otherwise do rotate
perturb.alpha=0.025  # scaling parameter for perturb proposal, usually between
                     # 0.02-0.10, use this to tune acceptance rate of tree moves.
ncores=4      # number of parallel cpu cores when using OpenMP

# Other settings where defaults are usually fine.
alpha=0.95  # for the tree
beta=2.0    # depth prior.
nu=3        # \
sigq=0.99   #  for the
sigqf=0.90  #  variance priors.
kfac=2      # /

# Make the prior datastructures
tree.pri=maketreeprior(numcuts,alpha,beta,m)
dat.pri=makedatprior(tree.pri,y,yf,nu,sigq,sigqf,sigfhat,kfac,rho)
mcmc.info=makemcmcinfo(X,Xf,y,yf,caldims,nd,thin,burn,
	      pbd=pbd,perturb.alpha=perturb.alpha,ncores=ncores)


#################################################################################
# Run the model
#################################################################################
fit=calibrate(dat.pri,tree.pri,mcmc.info)

# Predict the out-of-sample field
fitp=predict(Xp,mcmc.info,0.025,0.975)

#################################################################################
# Do some plots
#################################################################################
png("mcmc.png",width=900,height=600)
par(mfrow=c(2,3))
# In-sample fits
plot(yf,fit$yf,pch=20,cex=.5,xlab="Field Observations",ylab="Posterior Mean")
abline(0,1)
plot(y,fit$y,pch=20,cex=.5,xlab="Simulator Outputs", ylab="Posterior Mean")
abline(0,1)
# Trace plot of sigma,sigmaf
plot(fit$sd,type='l',xlab="Iteration",ylab=expression(sigma))
plot(fit$sdf,type='l',xlab="Iteration",ylab=expression(phi))
# Plot of calibration parameter estimates
plot(density(fit$thetas[,1]),xlim=c(-5,5),xlab="Parameter Estimates",main="",lwd=2)
lines(density(fit$thetas[,2]),col="blue",lwd=2)
abline(v=caltrues[1],lwd=.5)
mtext(expression(theta[1]),at=caltrues[1],side=1,cex=.7)
mtext(expression(theta[2]),at=caltrues[2],side=1,cex=.7,col="blue")
abline(v=caltrues[2],col="blue",lwd=.5)
dev.off()

# Predictions
ymax=max(7)
ymin=min(-5)
xmin=min(fp)
xmax=max(fp)

png("predict.png",width=1200,height=600)
par(mfrow=c(1,2))
plot(fp,fitp$y.5,pch=20,cex=1.5,xlab="True Field Process",ylab="Posterior Median Estimates and 95% credible intervals",ylim=c(ymin-1,ymax+1),xlim=c(xmin,xmax))
for(i in 1:np)
  lines(c(fp[i],fp[i]),c(fitp$y.l[i],fitp$y.u[i]),col="grey")
points(fp,fitp$y.5,pch=20,cex=1.5,col="grey")
for(i in 1:np)
  lines(c(fp[i],fp[i]),c(fitp$yf.l[i],fitp$yf.u[i]),col="green")
points(fp,fitp$yf.5,pch=20,cex=1.5,col="green")
for(i in 1:np)
  lines(c(fp[i],fp[i]),c(fitp$d.l[i],fitp$d.u[i]),col="orange")
points(fp,fitp$d.5,pch=20,cex=1.5,col="orange")
abline(0,1)
ix=sort(fp,index.return=TRUE)$ix
lines(fp[ix],bias[ix],lwd=1.5,lty=3)

ix=sort(Xp[,1],index.return=TRUE)$ix
plot(Xp[ix,1],fp[ix],type='l',lwd=1,col="black",xlab="x",ylab="Field Process",ylim=c(min(ymin,xmin),max(ymax,xmax)))
points(Xp[,1],fitp$y.5,pch=20,cex=1.5,col="grey")
for(i in 1:np)
  lines(c(Xp[i,1],Xp[i,1]),c(fitp$y.l[i],fitp$y.u[i]),col="grey")
points(Xp[,1],fitp$yf.5,pch=20,cex=1.5,col="green")
for(i in 1:np)
  lines(c(Xp[i,1],Xp[i,1]),c(fitp$yf.l[i],fitp$yf.u[i]),col="green")
points(Xp[,1],fitp$d.5,pch=20,cex=1.5,col="orange")
for(i in 1:np)
  lines(c(Xp[i,1],Xp[i,1]),c(fitp$d.l[i],fitp$d.u[i]),col="orange")
lines(Xp[ix,1],bias[ix],lwd=1,lty=3)
points(Xf[,1],yf,pch='*',cex=2,col="red")
dev.off()

png("identifiability.png",width=900,height=300)
par(mfrow=c(1,3))
meand=apply(fitp$postd,1,mean) #vector of the mean level of prediction surface for all posterior draws
cor(fit$thetas[,1],meand) # is the mean level correlated with one of the
cor(fit$thetas[,2],meand) # calibration parameters?

plot(fit$thetas[,1],meand,pch=20,cex=1,ylab="mean discrepancy level",xlab=expression(theta[1]),xlim=range(fit$thetas[,1]),ylim=c(min(meand),max(meand)),
     main=paste("correlation=",round(cor(fit$thetas[,1],meand),3),sep=""))
abline(h=mean(meand))
plot(fit$thetas[,2],meand,pch=20,cex=1,ylab="mean discrepancy level", xlab=expression(theta[2]),xlim=range(fit$thetas[,2]),ylim=c(min(meand),max(meand)),
     main=paste("correlation=",round(cor(fit$thetas[,2],meand),3),sep=""))
abline(h=mean(meand))

cormat=matrix(0,nrow=2,ncol=np)
for(i in 1:np) {
	cormat[1,i]=cor(fit$thetas[,1],fitp$postd[,i])
	cormat[2,i]=cor(fit$thetas[,2],fitp$postd[,i])
}
plot(Xp[,1],cormat[1,],pch=20,cex=1,ylim=c(-1,1),xlab=expression(x[1]),ylab="correlation",main=expression("black=theta[1]  blue=theta[2]"))
points(Xp[,1],cormat[2,],pch=20,cex=1,col="blue")
abline(h=0)
dev.off()

